package me.akito123321.valoStrats.rest.controller;

import java.util.List;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import me.akito123321.valoStrats.rest.requests.GroupRequest;
import me.akito123321.valoStrats.rest.requests.LineupRequest;
import me.akito123321.valoStrats.rest.requests.MapRequest;
import me.akito123321.valoStrats.rest.requests.PlayerTypeRequest;
import me.akito123321.valoStrats.rest.requests.StratRequest;
import me.akito123321.valoStrats.rest.requests.StratStateRequest;
import me.akito123321.valoStrats.rest.services.GroupService;
import me.akito123321.valoStrats.rest.services.LineupService;
import me.akito123321.valoStrats.rest.services.MapService;
import me.akito123321.valoStrats.rest.services.PlayerTypeService;
import me.akito123321.valoStrats.rest.services.ProfileService;
import me.akito123321.valoStrats.rest.services.StratService;
import me.akito123321.valoStrats.rest.services.StratStateService;
import me.akito123321.valoStrats.rest.util.JDBCUserDetailsService;
import me.akito123321.valoStrats.schemas.Group;
import me.akito123321.valoStrats.schemas.Lineup;
import me.akito123321.valoStrats.schemas.Map;
import me.akito123321.valoStrats.schemas.PlayerType;
import me.akito123321.valoStrats.schemas.Profile;
import me.akito123321.valoStrats.schemas.Strat;
import me.akito123321.valoStrats.schemas.StratState;
import me.akito123321.valoStrats.schemas.StratsUser;

@RestController
@RequestMapping("/api/group")
public class GroupRestAPI {
	
	@Autowired
	private PlayerTypeService playerTypeService;
	
	@Autowired
	private LineupService lineupService;
	
	@Autowired
	private ProfileService profileService;
	
	@Autowired
	private MapService mapService;
	
	@Autowired
	private StratStateService stratStateService;

	@Autowired
	private JDBCUserDetailsService userService;

	@Autowired
	private GroupService groupService;
	
	@Autowired
	private StratService stratService;

	@PostMapping()
	public ResponseEntity<Group> createGroup(@RequestBody GroupRequest request) {
		StratsUser user = getUser();
		Group group = new Group(request.name(), user);
		group.getMembers().add(user);
		groupService.saveGroup(group);
		return ResponseEntity.ok(group);
	}

	@GetMapping()
	public ResponseEntity<List<Group>> getGroups() {
		StratsUser user = getUser();
		List<Group> groups = user.getGroups();
		return ResponseEntity.ok(groups);
	}

	@DeleteMapping("/{id}")
	public ResponseEntity<Void> removeGroup(@PathVariable String id) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		if (group == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getOwner().getUsername().equals(user.getUsername())) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}

		groupService.removeGroup(id);

		return ResponseEntity.ok(null);
	}

	@PutMapping("/{id}")
	public ResponseEntity<Group> updateGroup(@PathVariable String id, @RequestBody GroupRequest request) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		if (group == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getOwner().getUsername().equals(user.getUsername())) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}

		group.setName(request.name());

		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}

	@GetMapping("/{id}")
	public ResponseEntity<Group> getGroup(@PathVariable String id) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		if (group == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}

		return ResponseEntity.ok(group);
	}
	
	@PostMapping("/{id}/strat-type")
	public ResponseEntity<Group> addStratType(@PathVariable String id, @RequestBody String typeName) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		if (group == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if(group.getStratTypes().contains(typeName)) {
			return ResponseEntity.status(HttpStatus.CONFLICT).build();
		}
		
		group.getStratTypes().add(typeName);

		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}
	
	@DeleteMapping("/{id}/strat-type/{typeName}")
	public ResponseEntity<Group> removeStratType(@PathVariable String id, @PathVariable String typeName) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		if (group == null | !group.getStratTypes().contains(typeName)) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		
		group.getStratTypes().remove(typeName);

		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}

	@PostMapping("/{id}/member")
	public ResponseEntity<Group> addUserToGroup(@PathVariable String id, @RequestBody String userId) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		if (group == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getOwner().equals(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}

		group.getMembers().add(userService.loadUserByUsername(userId));

		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}

	@DeleteMapping("/{id}/member/{userId}")
	public ResponseEntity<Group> removeUserFromGroup(@PathVariable String id, @PathVariable String userId) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		if (group == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getOwner().getUsername().equals(user.getUsername()) || !user.getUsername().equals(userId)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}

		group.getMembers().remove(userService.loadUserByUsername(userId));

		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}

	@PostMapping("/{id}/strat")
	public ResponseEntity<Group> addStrat(@PathVariable String id, @RequestBody StratRequest stratRequest) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		if (group == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}

		Strat strat = new Strat(stratRequest.title(), stratRequest.description(), group, stratRequest.stratType());
		
		if(stratRequest.attempts() != null) {
			strat.setAttempts(stratRequest.attempts());
		}
		if(stratRequest.success() != null) {
			strat.setSuccess(stratRequest.success());
		}
		if(!group.getStratTypes().contains(stratRequest.stratType())) {
			group.getStratTypes().add(stratRequest.stratType());
		}

		stratService.saveStrat(strat);
		
		group.getStrats().add(strat);
		
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}

	@DeleteMapping("/{id}/strat/{stratId}")
	public ResponseEntity<Group> removeStrat(@PathVariable String id, @PathVariable String stratId) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Strat strat = stratService.getStratById(stratId);
		if (group == null | strat == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if (!group.getStrats().contains(strat)) {
			return ResponseEntity.badRequest().build();
		}
		
		stratService.removeStrat(strat);

		group.getStrats().remove(strat);

		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}
	
	@PutMapping("/{id}/strat/{stratId}")
	public ResponseEntity<Group> editStrat(@PathVariable String id, @PathVariable String stratId, @RequestBody StratRequest newStrat) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Strat strat = stratService.getStratById(stratId);
		if (group == null | strat == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if (!group.getStrats().contains(strat)) {
			return ResponseEntity.badRequest().build();
		}
		
		strat.setDescription(newStrat.description());
		strat.setTitle(newStrat.title());
		strat.setStratStates(newStrat.stratStates());
		strat.setStratType(newStrat.stratType());
		if(newStrat.attempts() != null) {
			strat.setAttempts(newStrat.attempts());
		}
		if(newStrat.success() != null) {
			strat.setSuccess(newStrat.success());
		}
		if(!group.getStratTypes().contains(newStrat.stratType())) {
			group.getStratTypes().add(newStrat.stratType());
		}
		
		stratService.saveStrat(strat);
		
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}
	
	@PostMapping("/{id}/map")
	public ResponseEntity<Group> addMap(@PathVariable String id, @RequestBody MapRequest mapRequest) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		if (group == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}

		Map map = new Map(mapRequest.name(), mapRequest.image());

		mapService.saveMap(map);
		
		group.getMaps().add(map);
		
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}

	@DeleteMapping("/{id}/map/{mapId}")
	public ResponseEntity<Group> removeMap(@PathVariable String id, @PathVariable String mapId) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Map map = mapService.getMapById(mapId);
		if (group == null | map == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if (!group.getMaps().contains(map)) {
			return ResponseEntity.badRequest().build();
		}
		
		mapService.removeMap(map);

		group.getMaps().remove(map);

		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}
	
	@PutMapping("/{id}/map/{mapId}")
	public ResponseEntity<Group> editMap(@PathVariable String id, @PathVariable String mapId, @RequestBody MapRequest newMap) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Map map = mapService.getMapById(mapId);
		if (group == null | map == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if (!group.getMaps().contains(map)) {
			return ResponseEntity.badRequest().build();
		}
		
		map.setName(newMap.name());
		map.setImage(newMap.image());
		
		mapService.saveMap(map);
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}
	
	@PostMapping("/{id}/profile")
	public ResponseEntity<Group> addProfile(@PathVariable String id, @RequestBody ProfileRequest profileRequest) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		if (group == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}

		Profile profile = new Profile(profileRequest.name(), profileRequest.image());

		profileService.saveProfile(profile);
		
		group.getProfiles().add(profile);
		
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}

	@DeleteMapping("/{id}/profile/{profileId}")
	public ResponseEntity<Group> removeProfile(@PathVariable String id, @PathVariable String profileId) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Profile profile = profileService.getProfileById(profileId);
		if (group == null | profile == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if (!group.getProfiles().contains(profile)) {
			return ResponseEntity.badRequest().build();
		}
		
		profileService.removeProfile(profile);

		group.getProfiles().remove(profile);

		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}
	
	@PutMapping("/{id}/profile/{profileId}")
	public ResponseEntity<Group> editProfile(@PathVariable String id, @PathVariable String profileId, @RequestBody MapRequest newMap) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Profile profile = profileService.getProfileById(profileId);
		if (group == null | profile == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if (!group.getProfiles().contains(profile)) {
			return ResponseEntity.badRequest().build();
		}
		
		profile.setName(newMap.name());
		profile.setImage(newMap.image());
		
		profileService.saveProfile(profile);
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}
	
	@PostMapping("/{id}/strat/{stratId}/player-type")
	public ResponseEntity<Group> addPlayerType(@PathVariable String id, @PathVariable String stratId , @RequestBody PlayerTypeRequest playerTypeRequest) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Strat strat = stratService.getStratById(stratId);
		
		if (group == null | strat == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if(!group.getStrats().contains(strat)) {
			return ResponseEntity.badRequest().build();
		}
		
		PlayerType playerType = new PlayerType(playerTypeRequest.name(), playerTypeRequest.task());
		
		playerTypeService.savePlayerType(playerType);

		strat.getPlayerTypes().add(playerType);
		
		stratService.saveStrat(strat);
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}

	@DeleteMapping("/{id}/strat/{stratId}/player-type/{playerTypeId}")
	public ResponseEntity<Group> removePlayerType(@PathVariable String id, @PathVariable String stratId, @PathVariable String playerTypeId) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Strat strat = stratService.getStratById(stratId);
		PlayerType playerType = playerTypeService.getPlayerTypeById(playerTypeId);
		
		if (group == null | strat == null | playerType == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if(!group.getStrats().contains(strat) | !strat.getPlayerTypes().contains(playerType)) {
			return ResponseEntity.badRequest().build();
		}
		
		playerTypeService.removePlayerType(playerType);
		
		strat.getPlayerTypes().remove(playerType);
		
		stratService.saveStrat(strat);
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}
	
	@PutMapping("/{id}/strat/{stratId}/player-type/{playerTypeId}")
	public ResponseEntity<Group> editPlayerType(@PathVariable String id, @PathVariable String stratId, @PathVariable String playerTypeId, @RequestBody PlayerTypeRequest newPlayerType) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Strat strat = stratService.getStratById(stratId);
		PlayerType playerType = playerTypeService.getPlayerTypeById(playerTypeId);
		
		if (group == null | strat == null | playerType == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if(!group.getStrats().contains(strat) | !strat.getPlayerTypes().contains(playerType)) {
			return ResponseEntity.badRequest().build();
		}
		
		playerType.setName(newPlayerType.name());
		playerType.setTask(newPlayerType.task());
		
		playerTypeService.savePlayerType(playerType);
		stratService.saveStrat(strat);
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}
	
	@PostMapping("/{id}/strat/{stratId}/player-type/{playerTypeId}/profile")
	public ResponseEntity<Group> addProfile(@PathVariable String id, @PathVariable String stratId, @PathVariable String playerTypeId , @RequestBody ProfileRequest profileRequest) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Strat strat = stratService.getStratById(stratId);
		PlayerType playerType = playerTypeService.getPlayerTypeById(playerTypeId);
		
		if (group == null | strat == null| playerTypeId == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if(!group.getStrats().contains(strat) && !strat.getPlayerTypes().contains(playerType)) {
			return ResponseEntity.badRequest().build();
		}
		
		Profile profile = new Profile(profileRequest.name(), profileRequest.image());
		
		profileService.saveProfile(profile);

		playerType.getProfiles().add(profile);
		
		playerTypeService.savePlayerType(playerType);
		stratService.saveStrat(strat);
		
		if(!group.getProfiles().contains(profile)) {
			group.getProfiles().add(profile);
		}
		
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}

	@DeleteMapping("/{id}/strat/{stratId}/player-type/{playerTypeId}/profile/{profileId}")
	public ResponseEntity<Group> removeProfile(@PathVariable String id, @PathVariable String stratId, @PathVariable String playerTypeId, @PathVariable String profileId) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Strat strat = stratService.getStratById(stratId);
		PlayerType playerType = playerTypeService.getPlayerTypeById(playerTypeId);
		Profile profile = profileService.getProfileById(profileId);
		
		if (group == null | strat == null | playerType == null | profile == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if(!group.getStrats().contains(strat) | !strat.getPlayerTypes().contains(playerType) | !playerType.getProfiles().contains(profile)) {
			return ResponseEntity.badRequest().build();
		}
		
		profileService.removeProfile(profile);
		
		playerType.getProfiles().remove(profile);
		
		playerTypeService.savePlayerType(playerType);
		stratService.saveStrat(strat);
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}
	
	@PutMapping("/{id}/strat/{stratId}/player-type/{playerTypeId}profile/{profileId}")
	public ResponseEntity<Group> editProfile(@PathVariable String id, @PathVariable String stratId, @PathVariable String playerTypeId, @PathVariable String profileId, @RequestBody ProfileRequest newProfile) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Strat strat = stratService.getStratById(stratId);
		PlayerType playerType = playerTypeService.getPlayerTypeById(playerTypeId);
		Profile profile = profileService.getProfileById(profileId);
		
		if (group == null | strat == null | playerType == null | profile == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if(!group.getStrats().contains(strat) | !strat.getPlayerTypes().contains(playerType) | !playerType.getProfiles().contains(profile)) {
			return ResponseEntity.badRequest().build();
		}
		
		removeProfile(id, stratId, playerTypeId, profileId);
		addProfile(id, stratId, playerTypeId, newProfile);

		return ResponseEntity.ok(group);
	}
	
	@PostMapping("/{id}/strat/{stratId}/strat-state")
	public ResponseEntity<Group> addStratState(@PathVariable String id, @PathVariable String stratId , @RequestBody StratStateRequest stratStateRequest) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Strat strat = stratService.getStratById(stratId);
		
		if (group == null | strat == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if(!group.getStrats().contains(strat)) {
			return ResponseEntity.badRequest().build();
		}
		
		StratState stratState = new StratState(stratStateRequest.description(), strat);
		
		stratStateService.saveStratState(stratState);

		strat.getStratStates().add(stratState);
		
		stratService.saveStrat(strat);
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}

	@DeleteMapping("/{id}/strat/{stratId}/strat-state/{stratStateId}")
	public ResponseEntity<Group> removeStratState(@PathVariable String id, @PathVariable String stratId, @PathVariable String stratStateId) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Strat strat = stratService.getStratById(stratId);
		StratState stratState = stratStateService.getStratStateById(stratStateId);
		
		if (group == null | strat == null | stratState == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if(!group.getStrats().contains(strat) | !strat.getStratStates().contains(stratState)) {
			return ResponseEntity.badRequest().build();
		}
		
		stratStateService.removeStratState(stratState);
		
		strat.getStratStates().remove(stratState);
		
		stratService.saveStrat(strat);
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}
	
	@PutMapping("/{id}/strat/{stratId}/strat-state/{stratStateId}")
	public ResponseEntity<Group> editStratState(@PathVariable String id, @PathVariable String stratId, @PathVariable String stratStateId, @RequestBody StratStateRequest newStratState) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Strat strat = stratService.getStratById(stratId);
		StratState stratState = stratStateService.getStratStateById(stratStateId);
		
		if (group == null | strat == null | stratState == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if(!group.getStrats().contains(strat) | !strat.getStratStates().contains(stratState)) {
			return ResponseEntity.badRequest().build();
		}
		
		stratState.setDescription(newStratState.description());
		stratState.setImage(newStratState.image());
		
		stratStateService.saveStratState(stratState);
		stratService.saveStrat(strat);
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}
	
	@PostMapping("/{id}/strat/{stratId}/strat-state/{stratStateId}/lineup")
	public ResponseEntity<Group> addLineup(@PathVariable String id, @PathVariable String stratId, @PathVariable String stratStateId, @RequestBody LineupRequest lineupRequest) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Strat strat = stratService.getStratById(stratId);
		StratState stratState = stratStateService.getStratStateById(stratStateId);
		
		if (group == null | strat == null | stratState == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if(!group.getStrats().contains(strat) && !strat.getStratStates().contains(stratState)) {
			return ResponseEntity.badRequest().build();
		}
		
		Lineup lineup = new Lineup(lineupRequest.description(), lineupRequest.image());
		
		lineupService.saveLineup(lineup);

		stratState.getLineups().add(lineup);
		
		stratStateService.saveStratState(stratState);
		stratService.saveStrat(strat);
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}

	@DeleteMapping("/{id}/strat/{stratId}/strat-state/{stratStateId}/lineup/{lineupId}")
	public ResponseEntity<Group> removeLineup(@PathVariable String id, @PathVariable String stratId, @PathVariable String stratStateId, @PathVariable String linupId) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Strat strat = stratService.getStratById(stratId);
		StratState stratState = stratStateService.getStratStateById(stratStateId);
		Lineup lineup = lineupService.getLineupById(linupId);
		
		if (group == null | strat == null | stratState == null | lineup == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if(!group.getStrats().contains(strat) | !strat.getStratStates().contains(stratState) | !stratState.getLineups().contains(lineup)) {
			return ResponseEntity.badRequest().build();
		}
		
		lineupService.removeLineup(lineup);
		
		stratState.getLineups().remove(lineup);

		stratService.saveStrat(strat);
		stratService.saveStrat(strat);
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}
	
	@PutMapping("/{id}/strat/{stratId}/strat-state/{stratStateId}/lineup/{lineupId}")
	public ResponseEntity<Group> editLineup(@PathVariable String id, @PathVariable String stratId, @PathVariable String stratStateId, @PathVariable String lineupId, @RequestBody LineupRequest newLineup) {
		StratsUser user = getUser();

		Group group = groupService.getGroupById(id);
		Strat strat = stratService.getStratById(stratId);
		StratState stratState = stratStateService.getStratStateById(stratStateId);
		Lineup lineup = lineupService.getLineupById(lineupId);
		
		if (group == null | strat == null | stratState == null | lineup == null) {
			return ResponseEntity.notFound().build();
		}
		if (!group.getMembers().contains(user)) {
			return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
		}
		if(!group.getStrats().contains(strat) | !strat.getStratStates().contains(stratState) | !stratState.getLineups().contains(lineup)) {
			return ResponseEntity.badRequest().build();
		}
		
		lineup.setDescription(newLineup.description());
		lineup.setLineupImage(newLineup.image());
		
		lineupService.saveLineup(lineup);
		stratStateService.saveStratState(stratState);
		stratService.saveStrat(strat);
		groupService.saveGroup(group);

		return ResponseEntity.ok(group);
	}
	

	private StratsUser getUser() {
		Authentication auth = SecurityContextHolder.getContext().getAuthentication();
		StratsUser user = (StratsUser) auth.getPrincipal();
		return user;
	}

}
